// Copyright (c) 2005 David Drysdale.  See license.txt for license info
#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif

#include <PalmOS.h>

#include "builddate.h"
#include "board.h"
#include "utils.h"
#include "constants.h"

BOARD *CreateBoard(Int16 width, Int16 height, Int16 numMines)
{
  BOARD *board = NULL;
  UInt32 allocSize = sizeof(BOARD) + (width*height*sizeof(UInt8));

  // Allocate the structure
  board = MemPtrNew(allocSize);
  if (board == 0) goto EXIT;

  // Fill in the fields that are fixed for a given board
  board->length = (Int16) allocSize;
  StrCopy(board->buildDate, BUILDDATE);
  board->width = width;
  board->height = height;
  board->totalMines = numMines;

  // Initialize the best times
  board->preferences = 0;
  ResetTimes(board);
EXIT:
  return board;
}

void CopyPreferences(BOARD *dst, BOARD *src)
{
  dst->smallEasyTime = src->smallEasyTime;
  dst->smallMediumTime = src->smallMediumTime;
  dst->smallHardTime = src->smallHardTime;
  dst->mediumEasyTime = src->mediumEasyTime;
  dst->mediumMediumTime = src->mediumMediumTime;
  dst->mediumHardTime = src->mediumHardTime;
  dst->largeEasyTime = src->largeEasyTime;
  dst->largeMediumTime = src->largeMediumTime;
  dst->largeHardTime = src->largeHardTime;
  dst->preferences = src->preferences;
}

void CopyBoard(BOARD *dst, BOARD *src) 
{
  Int16 row;
  Int16 col;

  //DEBUG((T, "status %x mines %d squares %d", src->gameStatus, src->minesLeftToMark, src->unflaggedCellsToReveal));
  assert(src->width == dst->width);
  assert(src->height == dst->height);
  assert(src->length == dst->length);

  // Copy across best times and preferences
  CopyPreferences(dst, src);
  dst->totalMines = src->totalMines;
  dst->timeTaken = src->timeTaken;
  dst->gameStatus = src->gameStatus;
  dst->minesLeftToMark = src->minesLeftToMark;
  dst->unflaggedCellsToReveal = src->unflaggedCellsToReveal;

  for (row=0; row<src->height; row++) {
    for (col=0; col<src->width; col++) {
      CELLVALUE(dst, col, row) = CELLVALUE(src, col, row);
    }
  }
}

static void GenerateNeighborCounts(BOARD *board)
{
  Int16 row;
  Int16 col;

  // Figure out the number of neighbors for each non-mine
  for (row=0; row<board->height; row++) {
    for (col=0; col<board->width; col++) {
      CELLVALUE(board, col, row) &= ~NBRSMASK;
      if (!MINE(CELLVALUE(board, col, row))) {
        Int16 nbrRow;
        Int16 nbrCol;

        for (nbrRow=MAX(row-1,0); nbrRow < MIN(row+2, board->height); nbrRow++) {
          for (nbrCol=MAX(col-1,0); nbrCol < MIN(col+2, board->width); nbrCol++) {
            if ((nbrRow != row) || (nbrCol != col)) {
              if (MINE(CELLVALUE(board, nbrCol, nbrRow))) {
                CELLVALUE(board, col, row)++;
              }
            }
          }
        }
      }
    }
  }
}

void ResetTimes(BOARD *board)
{
  board->smallEasyTime = kTimeLimit;
  board->smallMediumTime = kTimeLimit;
  board->smallHardTime = kTimeLimit;
  board->mediumEasyTime = kTimeLimit;
  board->mediumMediumTime = kTimeLimit;
  board->mediumHardTime = kTimeLimit;
  board->largeEasyTime = kTimeLimit;
  board->largeMediumTime = kTimeLimit;
  board->largeHardTime = kTimeLimit;
}

void InitBoard(BOARD *board)
{
  Int16 count;
  Int16 row;
  Int16 col;
  UInt32 randRange = (UInt16)sysRandomMax + 1L;

  board->timeTaken = 0;
  board->gameStatus = 0x0000;
  board->minesLeftToMark = board->totalMines;
  board->unflaggedCellsToReveal = (board->width*board->height);

  // First clear out any old contents
  for (row=0; row<board->height; row++) {
    for (col=0; col<board->width; col++) {
      CELLVALUE(board, col, row) = 0x00;
    }
  }

  // Now randomly splatter the relevant number of mines
  count = board->totalMines;
  while (count > 0) {
    // Select row in range [0, height)
    row = (Int16)(((double)SysRandom(0)/(double)randRange)*board->height);
    // Select col in range [0, width)
    col = (Int16)(((double)SysRandom(0)/(double)randRange)*board->width);
    if (!MINE(CELLVALUE(board, col, row))) {
      // Found a space
      count--;
      CELLVALUE(board, col, row) = MINEMASK;
    }
  }

  GenerateNeighborCounts(board);
}

void MakeCellSafe(BOARD *board, Int16 col, Int16 row)
{
  Boolean done = false;
  Int16 moveRow;
  Int16 moveCol;
  UInt32 randRange = (UInt16)sysRandomMax + 1L;

  if (!MINE(CELLVALUE(board, col, row))) return;
  //DEBUG((T, "shuffle things so that (%d,%d) isn't a mine", col, row));
  while (!done) {
    // Select row in range [0, height)
    moveRow = (Int16)(((double)SysRandom(0)/(double)randRange)*board->height);
    // Select col in range [0, width)
    moveCol = (Int16)(((double)SysRandom(0)/(double)randRange)*board->width);
    if (!MINE(CELLVALUE(board, moveCol, moveRow))) {
      // Found a space - move the mine there 
      //DEBUG((T, "moving to (%d, %d)", moveCol, moveRow));
      CELLVALUE(board, moveCol, moveRow) = MINEMASK;
      CELLVALUE(board, col, row) &= ~MINEMASK;
      done = true;
    }
  }
  // Need to recalculate the numbers of neighbor everywhere
  GenerateNeighborCounts(board);  
}

void FlagCell(BOARD *board, Int16 col, Int16 row)
{
  if (REVEALED(CELLVALUE(board, col, row))) return;
  if (!FLAGGED(CELLVALUE(board, col, row))) {
    board->minesLeftToMark--;
    board->unflaggedCellsToReveal--;
    if ((board->unflaggedCellsToReveal == 0) && 
        (board->minesLeftToMark == 0)) {
      // All cells are either marked as mines, or revealed, and moreover
      // the number of mines flagged equals the actual number of mines => won game
      board->gameStatus |= GAMEWONMASK;
    }
  }
  CELLVALUE(board, col, row) |= FLAGMASK;
}
void QuestionCell(BOARD *board, Int16 col, Int16 row)
{
  if (REVEALED(CELLVALUE(board, col, row))) return;
  CELLVALUE(board, col, row) |= QUESTIONMASK;
}
void UnflagCell(BOARD *board, Int16 col, Int16 row)
{
  if (FLAGGED(CELLVALUE(board, col, row))) {
    board->minesLeftToMark++;
    board->unflaggedCellsToReveal++;
  }
  CELLVALUE(board, col, row) &= ~(FLAGMASK|QUESTIONMASK);
}
void Toggle2Cell(BOARD *board, Int16 col, Int16 row)
{
  if (REVEALED(CELLVALUE(board, col, row))) return;
  if (QUESTION(CELLVALUE(board, col, row))) {
    UnflagCell(board, col, row);
  } else if (FLAGGED(CELLVALUE(board, col, row))) {
    UnflagCell(board, col, row);
  } else {
    FlagCell(board, col, row);
  }
}
void Toggle3Cell(BOARD *board, Int16 col, Int16 row)
{
  if (REVEALED(CELLVALUE(board, col, row))) return;
  if (QUESTION(CELLVALUE(board, col, row))) {
    UnflagCell(board, col, row);
  } else if (FLAGGED(CELLVALUE(board, col, row))) {
    UnflagCell(board, col, row);
    QuestionCell(board, col, row);
  } else {
    FlagCell(board, col, row);
  }
}

static Int16 exposeCount;
static Int16 curCol;
static Int16 curRow;
static BOARD* curBoard;
void ExposeCellRecursive()
{
  UnflagCell(curBoard, curCol, curRow);
  if (!REVEALED(CELLVALUE(curBoard, curCol, curRow))) {
    curBoard->unflaggedCellsToReveal--;
    if ((curBoard->unflaggedCellsToReveal == 0) && 
        (curBoard->minesLeftToMark == 0)) {
      // All cells are either marked as mines, or revealed, and moreover
      // the number of mines flagged equals the actual number of mines => won game
      curBoard->gameStatus |= GAMEWONMASK;
    }
    exposeCount++;
    CELLVALUE(curBoard, curCol, curRow) |= REVEALEDMASK;
  }

  if (MINE(CELLVALUE(curBoard, curCol, curRow))) {
    curBoard->gameStatus |= GAMELOSTMASK;
  } else if (NBRS(CELLVALUE(curBoard, curCol, curRow)) == 0) {
    // This is a blank cell - go through the 8 neighbouring cells one by one,
    // automatically exposing them.
    
    if ((curRow > 0) && (curCol > 0) &&
        !REVEALED(CELLVALUE(curBoard, curCol-1, curRow-1))) {
      curCol--;
      curRow--;
      ExposeCellRecursive();
      curRow++;
      curCol++;
    }
    if ((curCol > 0) &&
        !REVEALED(CELLVALUE(curBoard, curCol-1, curRow))) {
      curCol--;
      ExposeCellRecursive();
      curCol++;
    }
    if ((curRow < (curBoard->height-1)) && (curCol > 0) &&
        !REVEALED(CELLVALUE(curBoard, curCol-1, curRow+1))) {
      curCol--;
      curRow++;
      ExposeCellRecursive();
      curRow--;
      curCol++;
    }


    if ((curRow > 0) && 
        !REVEALED(CELLVALUE(curBoard, curCol, curRow-1))) {
      curRow--;
      ExposeCellRecursive();
      curRow++;
    }
    if ((curRow < (curBoard->height-1)) && 
        !REVEALED(CELLVALUE(curBoard, curCol, curRow+1))) {
      curRow++;
      ExposeCellRecursive();
      curRow--;
    }

    if ((curRow > 0) && (curCol < (curBoard->width-1)) &&
        !REVEALED(CELLVALUE(curBoard, curCol+1, curRow-1))) {
      curCol++;
      curRow--;
      ExposeCellRecursive();
      curRow++;
      curCol--;
    }
    if ((curCol < (curBoard->width-1)) &&
        !REVEALED(CELLVALUE(curBoard, curCol+1, curRow))) {
      curCol++;
      ExposeCellRecursive();
      curCol--;
    }
    if ((curRow < (curBoard->height-1)) && (curCol < (curBoard->width-1)) &&
        !REVEALED(CELLVALUE(curBoard, curCol+1, curRow+1))) {
      curCol++;
      curRow++;
      ExposeCellRecursive();
      curRow--;
      curCol--;
    }
  }
}

Int16 ExposeCell(BOARD *board, Int16 col, Int16 row, Int16 *pExposeCount)
{

  // First exposure should never blow up a mine
  if (!(board->gameStatus & CELLEXPOSEDMASK)) {
    MakeCellSafe(board, col, row);
    board->gameStatus |= CELLEXPOSEDMASK;
  }

  // Put the parameters into static variables to reduce stack size for 
  // the recursion
  exposeCount = 0;
  curCol = col;
  curRow = row;
  curBoard = board;

  // Do the recursive exposure
  ExposeCellRecursive();

  *pExposeCount = exposeCount;
  return GAMELOST(board) ? 1 : 0;
}

Int16 DeduceCell(BOARD *board, Int16 col, Int16 row, Int16 *pExposeCount)
{
  Int16 nbrRow;
  Int16 nbrCol;
  Int16 flagCount = 0;
  Int16 nonFlagCount = 0;
  Int16 retval = 0;

  *pExposeCount = 0;

  // This operation is only applicable to revealed cells
  if (!REVEALED(CELLVALUE(board, col, row))) return 0;  

  // Count the number of flags in the surrounding area
  for (nbrRow=MAX(row-1,0); nbrRow < MIN(row+2, board->height); nbrRow++) {
    for (nbrCol=MAX(col-1,0); nbrCol < MIN(col+2, board->width); nbrCol++) {
      if ((row != nbrRow) || (col != nbrCol)) {
        if (!REVEALED(CELLVALUE(board, nbrCol, nbrRow))) {
          if (FLAGGED(CELLVALUE(board, nbrCol, nbrRow))) {
            flagCount++;
          } else {
            nonFlagCount++;
          }
        }
      }
    }
  }
  //DEBUG((T, "(%d,%d) has %d flags nearby, %d nonflags", col, row, flagCount, nonFlagCount));

  if (flagCount == NBRS(CELLVALUE(board, col, row))) {
    // All of the mines in the neighbourhood are allegedly accounted for, so 
    // expose all of the neighbourhood
    for (nbrRow=MAX(row-1,0); nbrRow < MIN(row+2, board->height); nbrRow++) {
      for (nbrCol=MAX(col-1,0); nbrCol < MIN(col+2, board->width); nbrCol++) {
        if ((row != nbrRow) || (col != nbrCol)) {
          if (!REVEALED(CELLVALUE(board, nbrCol, nbrRow)) &&
              !FLAGGED(CELLVALUE(board, nbrCol, nbrRow))) {
            retval |= ExposeCell(board, nbrCol, nbrRow, pExposeCount);
          }
        }
      }
    }
  } else if ((flagCount+nonFlagCount) == NBRS(CELLVALUE(board, col, row))) {
    // The flagged and unflagged covered cells completely account for all of the possible places a mine could be, so
    // automagically flag the remaining unrevealed cells
    for (nbrRow=MAX(row-1,0); nbrRow < MIN(row+2, board->height); nbrRow++) {
      for (nbrCol=MAX(col-1,0); nbrCol < MIN(col+2, board->width); nbrCol++) {
        if ((row != nbrRow) || (col != nbrCol)) {
          if (!REVEALED(CELLVALUE(board, nbrCol, nbrRow)) &&
              !FLAGGED(CELLVALUE(board, nbrCol, nbrRow))) {
            FlagCell(board, nbrCol, nbrRow);
          }
        }
      }
    }
  }
  return retval;
}

void ShowAllMines(BOARD *board)
{
  Int16 row;
  Int16 col;

  for (row=0; row<board->height; row++) {
    for (col=0; col<board->width; col++) {
      if (MINE(CELLVALUE(board, col, row))) {
        UnflagCell(board, col, row);
        CELLVALUE(board, col, row) |= REVEALEDMASK;
      } else if (FLAGGED(CELLVALUE(board, col, row))) {
        CELLVALUE(board, col, row) |= REVEALEDMASK;
      }
    }
  }
}
